#include <stdio.h>
#include <stdlib.h>
#include <sys/ioctl.h>
#include <fcntl.h>
#include <errno.h>
#include <string.h>
#include <unistd.h>
#include "snerdConfig.h"
#include "snerdDB.h"
#include "bergenUDPSocket.h"
#include "SampleFile.h"
#include "RATSource.h"
#include "RecordFile.h"
#include "Tone.h"
#include "WhiteNoise.h"
#include "Amplitude.h"
#include "Listener.h"
#include "AudioInterface.h"
#include "snerdServer.h"

// Setting the default values
snerdDB * snerdServer::db_ = NULL;
bergenUDPSocket * snerdServer::rcvSocket_ = NULL;
bergenUDPSocket * snerdServer::sndSocket_ = NULL;
bool snerdServer::looping_ = false;
ConfigInformation* snerdServer::config_ = NULL;

// snerdServer Constructor initialises snerdServer object with given 
// buffer size, sample rate and the network connection socket 
void snerdServer::init(ConfigInformation* config)
	{
	db_ = new snerdDB(AudioInterface::bufferSize());
	config_ = config;
	rcvSocket_ = new bergenUDPSocket(config_->udpPort);
	rcvSocket_->setBlocking(0);
	}


// Once called the mainLoop functions only ends when the snerd server attribute
// "looping_" is disabled.
// With evry loop new arriving messages are processed, the sound objects are
// updated and the updated and mixed audiobuffer is written to the audiodevice
void snerdServer::mainLoop(void)
	{
	looping_ = true;
	while (looping_)
		{
		processMessages();
		AudioInterface::fillInputPort();
		db_->update();
		AudioInterface::writeOutputPort(db_->buffer(),db_->bufferSize());
		}
	delete rcvSocket_;
	delete sndSocket_;
	}


// If a message was received over the socket, the message is send to 
// splitMessage for getting messages isolated. The isolated massages are
// parsed by sending them to parseMessage if they are no sound related ones.
// Sound messages are forwarded to the according sound object
void snerdServer::processMessages(void)
	{
	char message[1024], *msgStrings[32];
	int numStrings;
	while (rcvSocket_->receive(message,sizeof(message)) > 0)
		{
		numStrings = splitMessage(message, msgStrings, 32);
		if (numStrings <= 0)
			continue;
		int handle = atoi(msgStrings[0]);
		if (handle == -1) //if something new is to be created
			parseMessage(msgStrings,numStrings);
		else  //if operating on something that does exist
			{
			//if operation for listener
			if((!strcasecmp(msgStrings[1],"setListenerPosition")) ||
			   (!strcasecmp(msgStrings[1],"killListener")) )
				{
				Listener * listener = db_->findListener(handle);
				if (listener)
					listener->message(msgStrings+1,numStrings-1);
				else
					printf("WARNING: failed to find listener with handle %d\n",handle);
				}
			else
				{
				ASound * sound = db_->findHandle(handle);
				if (sound)
					sound->message(msgStrings+1,numStrings-1);
				else
					printf("WARNING: failed to find sound with handle %d (message:%s)\n",handle,msgStrings[1]);
				}
			}
		while (numStrings)
			{
			numStrings--;
			free(msgStrings[numStrings]);
			}
		}
	}


// The one message string is split into isolated messages. Each message is
// either seperated by \t \n or ' '. The amount of messages found is returned
int snerdServer::splitMessage(char *message,char **strings,int max)
	{
	int i=0, start, len, count=0;
	while ((message[i]) && (count < max))
		{
		i += strspn(message+i, " \t\n");
		if (!message[i])
			break;
		start = i;
		len = strcspn(message+start, " \t\n");
		strings[count] = (char *) malloc(len+1);
		strncpy(strings[count],message+start,len);
		strings[count][len] = '\0';
		count++;
		i += len;
		}
	return count;
	}


// Parses the message string and makes according function calls 
void snerdServer::parseMessage(char **message,int numStrings)
	{
	if (numStrings < 2)
		return;
	if (!strcasecmp(message[1],"new"))
		parseNewSoundMessage(message,numStrings);
	else if (!strcasecmp(message[1],"gain"))
		{
		if (numStrings > 2)
			db_->setGain(atof(message[2]));
		}
	else if (!strcasecmp(message[1],"path"))
		{
		if (numStrings > 2)
			db_->addPath(message[2]);
		}
	else if (!strcasecmp(message[1],"kill"))
		{
		looping_ = false;
		}
	else if (!strcasecmp(message[1],"reset"))
		{
		db_->reset();
		}
	else if (!strcasecmp(message[1],"ping"))
		{
		char* reply = "-1 pong";
		sndSocket_ = new bergenUDPSocket(0);
		sndSocket_->setDestination(rcvSocket_->lastFromAddress(),config_->udpPort+1);
		sndSocket_->setBlocking(0);
		sendMessage(reply,strlen(reply)+1);
		}
	else
		fprintf(stderr,"ERROR: unknown message \"%s\"\n",message[1]);
	}


// Parses the message for sound objects newly to generate 
void snerdServer::parseNewSoundMessage(char **message,int numStrings)
	{
	if (numStrings < 3)
		return;
	if (!strcasecmp(message[2],"listener"))
		{
		Listener* listener;
		listener = new Listener(config_,rcvSocket_->lastFromAddress());
		db_->addListener(listener);
		}
	else
		{
		ASound * sound = NULL;
		if (!strcasecmp(message[2],"sample") && numStrings>3)
			{
			if (numStrings > 4)
				sound = new SampleFile(message[3],db_,atoi(message[4]),config_);
			else
				sound = new SampleFile("",db_,atoi(message[3]),config_);
			sound->setSampleRate(AudioInterface::outputRate());
			}
		else if (!strcasecmp(message[2],"recordfile") && numStrings>3)
			{
			if (numStrings > 4)
				sound = new RecordFile(message[3],db_,atoi(message[4]),config_);
			else
				sound = new RecordFile("",db_,atoi(message[3]),config_);
			sound->setSampleRate(AudioInterface::inputRate());
			}
		else if (!strcasecmp(message[2],"ratsource") && numStrings>3)
			if (numStrings > 4)
				sound = new RATSource(message[3],db_,atoi(message[4]),config_);
			else
				sound = new RATSource("",db_,atoi(message[3]),config_);
		else if (!strcasecmp(message[2],"tone") && numStrings>3)
			{
			sound = new Tone(db_,atoi(message[3]),config_);
			sound->setSampleRate(AudioInterface::outputRate());
			}
		else if (!strcasecmp(message[2],"whitenoise") && numStrings>3)
			{
			sound = new WhiteNoise(db_,atoi(message[3]),config_);
			sound->setSampleRate(AudioInterface::outputRate());
			}
		else if (!strcasecmp(message[2],"amplitude") && numStrings>3)
			{
			sound = new Amplitude(db_,atoi(message[3]),config_);
			sound->setSampleRate(AudioInterface::inputRate());
			}
		else
			fprintf(stderr,"ERROR: asked to create unknown sound type '%s'\n",message[2]);
		if (sound)
			db_->addSound(sound);
		}	
	}


// Sends the message to the given listener
void snerdServer::sendMessage(void *msg,int len,int listener)
	{
	if (listener > 0)
		{
		Listener* list = db_->findListener(listener);
		if (list)
			sndSocket_->setDestination(list->address());
		}
	sndSocket_->send(msg,len);
	}


// Sends the message to the given client
void snerdServer::sendMessage(void *msg,int len,char* address)
	{
	if (address)
		sndSocket_->setDestination(address);
	sndSocket_->send(msg,len);
	}
